import shutil


def extract_nodes(sample, infile, minlen):
    output_links = f"gplas_input/{sample}_raw_links.txt"
    output_nodes = f"gplas_input/{sample}_raw_nodes.fasta"
    output_contigs = f"gplas_input/{sample}_contigs.fasta"

    with open(infile,'r') as graph, open(output_links,'w') as links, open(output_nodes,'w') as nodes, open(output_contigs,'w') as contigs:
        for line in graph:
            line = line.rstrip()
            if line[0] == 'S':
                cols = line.split('\t')
                number = cols[0] + str(cols[1])
                sequence = cols[2]
                if 'LN' in str(cols[3]): # Unicycler assembly
                    information = '_'.join(cols[3:])
                elif 'KC' in str(cols[3]): # Spades assembly
                    information = cols[3]
                else: # Empty sequence field (0 length), or other error in gfa format
                      # TODO exit the tool if the assembly is not from Unicycler or Spades? right now it would just skip all nodes and probably run into an error later in the workflow
                    continue # Skip node and continue to the next
                nodes.write(f">{number}_{information}\n{sequence}\n")
                if len(sequence) >= minlen:
                    contigs.write(f">{number}_{information}\n{sequence}\n")
            elif line[0] == 'L':
                links.write(f"{line}\n")


def extract_unbinned_solutions(sample):
    normal_results = f"results/normal_mode/{sample}_results_no_repeats.tab"
    bold_walks = f"walks/bold_mode/{sample}_solutions_bold.tab"
    unbinned_walks = f"walks/unbinned_nodes/{sample}_solutions_unbinned.tab"   
    #Get unbinned nodes
    unbinned_nodes = []
    with open(normal_results,'r') as file:
        for line in file:
            line = line.rstrip()
            cols = line.split('\t')
            component = cols[7]
            if component == 'Unbinned':
                number = str(cols[4])
                unbinned_nodes.append(number)
    #Select bold walks that start with unbinned nodes
    with open(bold_walks,'r') as infile, open(unbinned_walks,'w') as outfile:
        for line in infile:
            first_node = str(line.split('\t', 1)[0])
            first_node_unsigned = first_node.replace('+','')
            first_node_unsigned = first_node_unsigned.replace('-','')
            if first_node_unsigned in unbinned_nodes:
                outfile.write(line)
    #Combine solutions from bold and normal mode
    normal_walks = f"walks/normal_mode/{sample}_solutions.tab"
    combined_walks = f"walks/{sample}_solutions.tab"
    shutil.copyfile(normal_walks, combined_walks)
    with open(unbinned_walks,'r') as infile, open(combined_walks,'a') as outfile:
        for line in infile:
            outfile.write(line)



import pandas as pd
import numpy as np
from Bio.SeqIO.FastaIO import SimpleFastaParser


def scalar1(x):  # TODO can we do something like "from coocurrence.py import scalar1"? it is the same identical function
    denominator = (sum([value*value for value in x]))**0.5
    scaled_x = [value/denominator for value in x]
    return scaled_x


#TODO we have multiple loops with "for row in range(solutions.shape[0]):" can we possibly merge some of them?

def calculate_coocurrence_repeats(sample):
    #Inputs
    path_nodes = f"gplas_input/{sample}_raw_nodes.fasta"
    path_prediction = f"coverage/{sample}_clean_prediction.tab"
    input_solutions = f"walks/repeats/{sample}_solutions.tab"
    path_bins = f"results/{sample}_results_no_repeats.tab"
    clean_repeats_path = f"coverage/{sample}_clean_repeats.tab"

    #Outputs
    output_dir = "results/"
    output_results = f"results/{sample}_results.tab"
    output_components = f"results/{sample}_bins.tab"
    output_chromosomes = f"results/{sample}_chromosome_repeats.tab"

    clean_pred = pd.read_csv(path_prediction, sep='\t', header=0)
    clean_pred = clean_pred.astype({'Prob_Chromosome':float,
                                    'Prob_Plasmid':float,
                                    'Prediction':str,
                                    'Contig_name':str,
                                    'Contig_length':int,
                                    'number':str,
                                    'length':int,
                                    'coverage':float})

    solutions = pd.read_csv(input_solutions, sep='\t', header=None, names=['walks', 'initial_classification', 'unitig_classification', 'path_coverage'])
    max_nodes = max([walk.count(',')+1 for walk in solutions.loc[:,'walks']])
    steps = [''.join(['step_', str(step)]) for step in range(max_nodes)]
    stepsDf = pd.DataFrame(solutions.loc[:,'walks'].str.split(',', expand=True))
    stepsDf.columns = steps
    solutions = pd.concat([solutions, stepsDf], axis=1)

    solutions = solutions.drop(columns='walks')
    col_order = solutions.columns.to_list()
    col_order = col_order[3:] + col_order[:3]
    solutions = solutions[col_order]

    #get the last node of each solution
    last_nodes = []
    for row in range(solutions.shape[0]):
        last_node = [node for node in solutions.iloc[row,0:max_nodes].dropna()][-1]
        last_nodes.append(last_node)

    last_nodes_signless = [node.replace('+','') for node in last_nodes]
    last_nodes_signless = [node.replace('-','') for node in last_nodes_signless]

    solutions.loc[:,'last_nodes'] = last_nodes
    solutions.loc[:,'last_nodes_signless'] = last_nodes_signless

    #====Merge with bin data=============
    #Import information from the bins and match it with total pairs
    bins_data = pd.read_csv(path_bins, sep='\t', header=0)
    bins_data = bins_data.astype({'Prob_Chromosome':float,
                                  'Prob_Plasmid':float,
                                  'Prediction':str,
                                  'Contig_name':str,
                                  'number':str,
                                  'length':int,
                                  'coverage':float,
                                  'Bin':str})
    #merge it
    solutions = solutions.merge(bins_data.loc[:,['number','Bin']], how='left', left_on='last_nodes_signless', right_on='number')
    solutions = solutions.drop(columns='number')
    #assign 'C' to the chromosome and -1 to repeats
    #TODO ASK Julian: repeats are not assigned -1 in the original R-script? (i.e. comment above is lying) does it matter?
    index = solutions.loc[:,'Bin'].isna()
    solutions.loc[index,'Bin'] = 'C'

    #remove connections to unbinned unitigs
    index = solutions.loc[:,'Bin'] != 'Unbinned'
    solutions = solutions.loc[index,:]

    #remove connections to repeats only
    index = solutions.loc[:,'unitig_classification'] != 'Repeat'
    solutions = solutions.loc[index,:]

    #get the initial node without a symbol
    initial_nodes = [node.replace('+','') for node in solutions.loc[:,'step_0']]
    initial_nodes = [node.replace('-','') for node in initial_nodes]
    solutions.loc[:,'initial_node'] = initial_nodes
    solutions.reset_index(inplace=True, drop=True)

    #combine inital nodes with bin in a single variable
    solutions.loc[:,'repeat_bin'] = ['-'.join([solutions.loc[row,'initial_node'], solutions.loc[row,'Bin']]) for row in range(solutions.shape[0])]

    #keep cases in which the last node and second node are the same (repeat directly connected to a unitig)
    solutions.loc[:,'keep'] = ['yes' if ((solutions.loc[row,'step_1'] == solutions.loc[row,'last_nodes']) &
                                         (solutions.loc[row,'unitig_classification'] != 'Repeat')) else 'no' for row in range(solutions.shape[0])]

    index = solutions.loc[:,'keep'] == 'yes'
    final_solutions = solutions.loc[index,:]

    #----Analyze the remainig cases to check if we have the same bin upstream and downstream from the repeat--
    #separate into positive and negative solutions
    index = [solutions.loc[row,'step_1'][-1] == '+' for row in range(solutions.shape[0])]
    positive_solutions = solutions.loc[index,:]
    index = [solutions.loc[row,'step_1'][-1] == '-' for row in range(solutions.shape[0])]
    negative_solutions = solutions.loc[index,:]

    #Keep only the walks if they lead to same bin, or to the chromosome in both directions
    index = ((positive_solutions.loc[:,'repeat_bin'].isin(negative_solutions.loc[:,'repeat_bin'])) &
             (positive_solutions.loc[:,'keep'] == 'no'))
    positive_solutions_keep = positive_solutions.loc[index,:]
    index = ((negative_solutions.loc[:,'repeat_bin'].isin(positive_solutions.loc[:,'repeat_bin'])) &
             (negative_solutions.loc[:,'keep'] == 'no'))
    negative_solutions_keep = negative_solutions.loc[index,:]

    ## Filter only the valid walks
    final_walks = final_solutions.iloc[:,:max_nodes]
    final_positive_walks = positive_solutions_keep.iloc[:,:max_nodes]
    final_negative_walks = negative_solutions_keep.iloc[:,:max_nodes]

    solutions = pd.concat([final_walks, final_positive_walks, final_negative_walks], ignore_index=True)

    #create a list with all the nodes that appear in the plasmid walks.
    all_nodes = []
    for row in range(solutions.shape[0]):
        nodes = [node for node in solutions.loc[row,:].dropna()]
        all_nodes.extend(nodes)

    #get unique set of nodes
    unique_nodes = sorted(list(set(all_nodes)))

    #CREATE A CO-OCURRENCE MATRIX
    ##Column names are the nodes included in plasmid walks.
    ##Each row is a new walk
    ##Assign True if the node is present in walk and False if node is not present
    co_ocurrence = []
    for row in range(solutions.shape[0]):
        walk = [node for node in solutions.loc[row,:].dropna()]
        presence_absence = [node in walk for node in unique_nodes]
        co_ocurrence.append(presence_absence)
    co_ocurrence = pd.DataFrame(co_ocurrence, columns=unique_nodes)

    starting_nodes = [node for node in unique_nodes if node in solutions.loc[:,'step_0'].values]

    #create a dataframe for co-ocurrence frequency (in network format)
    #Start_node, Connecting_node, nr_occurences
    total_pairs = []
    #Get the number of times that two nodes co-ocuur in every walk
    for node in starting_nodes:
        index_col = [col_name == node for col_name in unique_nodes] # select the column of the target node
        index_walks = co_ocurrence.loc[:,index_col].values # select all walks where the target node is present
        walks = co_ocurrence.loc[index_walks,:].copy()
        col_sums = [sum(walks.loc[:,col]) for col in walks] # count how often any node is present in all selected walks
        for index_connecting_node in range(len(col_sums)): # save coocurrence data for every node/connecting_node
            connecting_node = unique_nodes[index_connecting_node]
            if connecting_node != node:
                weight = col_sums[index_connecting_node]
                total_pairs.append([node, connecting_node, weight])

    total_pairs = pd.DataFrame(total_pairs, columns=['Starting_node', 'Connecting_node', 'weight'])

    total_pairs.loc[:,'Starting_node'] = [node.replace('+','') for node in total_pairs.loc[:,'Starting_node']]
    total_pairs.loc[:,'Starting_node'] = [node.replace('-','') for node in total_pairs.loc[:,'Starting_node']]
    total_pairs.loc[:,'Connecting_node'] = [node.replace('+','') for node in total_pairs.loc[:,'Connecting_node']]
    total_pairs.loc[:,'Connecting_node'] = [node.replace('-','') for node in total_pairs.loc[:,'Connecting_node']]

    #Filter-out cases of no-coocurrence
    #TODO ASK Julian: filter is on weight > 1 why not weight > 0?
    index = [weight > 1 for weight in total_pairs.loc[:,'weight']]
    total_pairs = total_pairs.loc[index,:]

    #Scale weights
    complete_node_info = pd.DataFrame(columns=['Starting_node', 'Connecting_node', 'weight', 'scaled_weight'])
    for node in sorted(list(set(total_pairs.loc[:,'Starting_node']))):
        index = total_pairs.loc[:,'Starting_node'] == node
        first_node = total_pairs.loc[index,:]
        particular_node = []

        for connecting_node in sorted(list(set(first_node.loc[:,'Connecting_node']))):
            index = first_node.loc[:,'Connecting_node'] == connecting_node
            first_second_nodes = first_node.loc[index,:]
            total_weight = sum(first_second_nodes.loc[:,'weight'])
            particular_node.append([node, connecting_node, total_weight])

        particular_node = pd.DataFrame(particular_node, columns=['Starting_node', 'Connecting_node', 'weight'])
        particular_node.loc[:,'scaled_weight'] = scalar1(particular_node.loc[:,'weight']) # add a column with scaled weights using scalar1()
        if complete_node_info.shape[0] == 0:  # if statement to prevent "FutureWarning: concatenation with empty DataFrame"
            complete_node_info = particular_node
        else:
            complete_node_info = pd.concat([complete_node_info, particular_node], ignore_index=True)

    total_pairs = complete_node_info

    initial_nodes = [node.replace('+','') for node in starting_nodes]
    initial_nodes = [node.replace('-','') for node in initial_nodes]

    #get a list of clean untigs
    clean_unitigs = clean_pred.loc[:,'number']

    #Filter out connected repeated elements. Keep only connections from starting nodes (repeats) unitigs.
    index = ((total_pairs.loc[:,'Starting_node'].isin(initial_nodes)) &
             (total_pairs.loc[:,'Connecting_node'].isin(clean_unitigs)))
    total_pairs = total_pairs.loc[index,:]

    total_pairs = total_pairs.merge(bins_data.loc[:,['number', 'Bin']], how='left', left_on='Connecting_node', right_on='number')
    total_pairs = total_pairs.drop(columns='number')
    #ASSIGN 'C' to the chromosome
    index = total_pairs.loc[:,'Bin'].isna()
    total_pairs.loc[index,'Bin'] = 'C'

    total_pairs = total_pairs.loc[:,['Starting_node', 'Bin', 'weight', 'scaled_weight']]

    #===Reformat the dataframe to obtain the totality of co-courences
    #First check if we actually have co-ocurrence of unitigs.
    if total_pairs.shape[0] > 0 and total_pairs.shape[1] > 0:
        total_pairs.loc[:,'Pair'] = ['-'.join([total_pairs.loc[row,'Bin'], total_pairs.loc[row,'Starting_node']]) for row in range(total_pairs.shape[0])]
    else:
        return False

    single_edge_counting = []
    for pair in sorted(list(set(total_pairs.loc[:,'Pair']))):
        index = total_pairs.loc[:,'Pair'] == pair
        pairs_subset = total_pairs.loc[index,:]
        sum_weight = sum(pairs_subset.loc[:,'weight'])
        single_edge_counting.append([pair, sum_weight])

    pairs = [pair[0].split('-') for pair in single_edge_counting]

    weight_graph = pd.DataFrame(data={'From_to':[pair[1] for pair in pairs],
                                      'To_from':[pair[0] for pair in pairs],
                                      'weight':[weight[1] for weight in single_edge_counting]})

    #Get the data from coverages. Repeats and bins
    clean_repeats = pd.read_csv(clean_repeats_path, sep='\t', header=0)
    clean_repeats = clean_repeats.astype({'number':str,
                                          'coverage':float})

    bins_coverage = []
    for current_bin in sorted(list(set(bins_data.loc[:,'Bin']))):
        index = bins_data.loc[:,'Bin'] == current_bin
        bins_subset = bins_data.loc[index,:]
        mean_coverage = round(np.mean(bins_subset.loc[:,'coverage']), 2)
        bins_coverage.append([current_bin, mean_coverage])

    bins_coverage = pd.DataFrame(data=bins_coverage, columns=['Bin', 'bin_coverage'])

    weight_graph = weight_graph.merge(clean_repeats.loc[:,['number', 'coverage']], how='left', left_on='From_to', right_on='number')
    #TODO find a more elegant/efficient way to properly merge dataframes
    weight_graph = weight_graph.drop(columns='number')

    weight_graph = weight_graph.merge(bins_coverage, how='left', left_on='To_from', right_on='Bin')
    weight_graph = weight_graph.drop(columns='Bin')

    #assign a coverage of 1 to chromosomal unitigs
    index = weight_graph.loc[:,'bin_coverage'].isna()
    weight_graph.loc[index,'bin_coverage'] = float(1)

    #Explore if the combination of bins proposed by the algorithm is plausible based on coverage
    repeat_assignments = pd.DataFrame(columns=['From_to', 'To_from', 'weight', 'coverage', 'bin_coverage', 'rank'])
    #loop through each of the repeats
    for node in sorted(list(set(weight_graph.loc[:,'From_to']))):
        index = weight_graph.loc[:,'From_to'] == node
        df_node = weight_graph.loc[index,:].copy()
        #create a rank of the most likely connections, based on the co-ocurence count
        df_node.sort_values(by='weight', axis=0, ascending=False, ignore_index=True, inplace=True)
        df_node.loc[:,'rank'] = df_node.index.tolist()

        rank = 0 #start from the highest ranking (0)
        accumulated_cov = 0
        while rank <= max(df_node.loc[:,'rank']):
            index = df_node.loc[:,'rank'] == rank
            repeat_bin = df_node.loc[index,:]
            if repeat_bin.loc[:,'coverage'].values[0] + accumulated_cov >= repeat_bin.loc[:,'bin_coverage'].values[0]:
                accumulated_cov += repeat_bin.loc[:,'bin_coverage'].values[0]
                if repeat_assignments.shape[0] == 0: # if statement to prevent "FutureWarning: concatenation with empty DataFrame"
                    repeat_assignments = repeat_bin
                else:
                    repeat_assignments = pd.concat([repeat_assignments, repeat_bin], ignore_index=True)
            rank += 1
            
    repeat_assignments = repeat_assignments.loc[:,['To_from','From_to']].rename(columns={'To_from':'Bin',
                                                                                         'From_to':'number'})
    #separate results into plasmid and chromosome repeats
    index = repeat_assignments.loc[:,'Bin'] == 'C'
    chromosome_repeats = repeat_assignments.loc[index,:]

    index = repeat_assignments.loc[:,'Bin'] != 'C'
    plasmid_repeats = repeat_assignments.loc[index,:]

    print("Adding repeated elements to the predictions........... completed!")
    if plasmid_repeats.shape[0] == 0:
        print("gplas did not find repeated elements associated with plasmid predictions")
        bins_data.loc[:,'Prob_Chromosome'] = round(bins_data.loc[:,'Prob_Chromosome'], 2)
        bins_data.loc[:,'Prob_Plasmid'] = round(bins_data.loc[:,'Prob_Plasmid'], 2)
        bins_data.loc[:,'coverage'] = round(bins_data.loc[:,'coverage'], 2)
        full_info_assigned = bins_data
    else:
        print("gplas found repeated elements associated with plasmid predictions")
        #Get all the repeat nodes
        index = clean_repeats.loc[:,'number'].isin(plasmid_repeats.loc[:,'number']) # Selecting only contigs predicted as plasmid-derived
        pl_nodes = clean_repeats.loc[index,:]

        #Get all the information from the plasmid nodes (Length, coverage, bin number, etc)
        full_info_assigned = pl_nodes.merge(plasmid_repeats, on='number')

        #Add information to the results file
        full_info_assigned = full_info_assigned.drop(columns='Contig_length')
        full_info_assigned.loc[:,'Prob_Chromosome'] = round(full_info_assigned.loc[:,'Prob_Chromosome'], 2)
        full_info_assigned.loc[:,'Prob_Plasmid'] = round(full_info_assigned.loc[:,'Prob_Plasmid'], 2)
        full_info_assigned.loc[:,'coverage'] = round(full_info_assigned.loc[:,'coverage'], 2)
        full_info_assigned.loc[:,'Prediction'] = 'Repeat'

        bins_data.loc[:,'Prob_Chromosome'] = round(bins_data.loc[:,'Prob_Chromosome'], 2)
        bins_data.loc[:,'Prob_Plasmid'] = round(bins_data.loc[:,'Prob_Plasmid'], 2)
        bins_data.loc[:,'coverage'] = round(bins_data.loc[:,'coverage'], 2)

        full_info_assigned = pd.concat([full_info_assigned, bins_data], ignore_index=True)

    #Create the fasta files
    with open(path_nodes) as file:
        raw_nodes = [[str(values[0]), str(values[1])] for values in SimpleFastaParser(file)]

    df_nodes = pd.DataFrame(data=raw_nodes, columns=['Contig_name', 'Sequence'])
    df_nodes = df_nodes.merge(full_info_assigned, on='Contig_name')

    #Write fasta files
    for component in sorted(list(set(df_nodes.loc[:,'Bin']))):
        index = df_nodes.loc[:,'Bin'] == component
        nodes_component = df_nodes.loc[index,:]
        component_complete_name = f"{sample}_bin_{component}"
        filename = f"{output_dir}{component_complete_name}.fasta"

        with open(filename, mode='w') as file:
            for contig in range(nodes_component.shape[0]):
                file.write('>' + nodes_component.iloc[contig,0] + '\n' + nodes_component.iloc[contig,1] + '\n')

    results_summary = df_nodes.loc[:,['number', 'Bin']]

    full_info_assigned.to_csv(output_results, sep='\t', index=False, header=True, mode='w')
    results_summary.to_csv(output_components, sep='\t', index=False, header=True, mode='w')

    #format chromosome repeats and print
    chromosome_repeats = chromosome_repeats.loc[:,['number', 'Bin']] #TODO if possible change column order from the start instead of changing it at the end
    chromosome_repeats.loc[:,'Bin'] = 'Chromosome' #TODO if possible call the column name 'Chromosome' from the start instead of 'C' and renaming it at the end
    chromosome_repeats.to_csv(output_chromosomes, sep='\t', index=False, header=True, mode='w')

    return True



import argparse
import os
import sys
from plasmidCC.scripts.utils import speciesopts


def quit_tool(exitcode=0):
    if exitcode != 0:
        print('\n', end='')
        print("This run of gplas has ended unexpectedly. Please check above for any error messages")
        sys.exit(-1)
    else:
        sys.exit(0)


def file_exists(arg):
    if not os.path.isfile(arg):
        raise argparse.ArgumentTypeError(f"'{arg}' is not an existing file" + "\nPlease make sure the file exists and is spelled correctly")
    return str(arg)


def is_valid_file(arg, extensions=['gfa']):
    if not os.path.isfile(arg):
        raise argparse.ArgumentTypeError(f"'{arg}' is not an existing file" + "\nPlease make sure the file exists and is spelled correctly")
    _, file_extension = os.path.splitext(arg)
    if not file_extension[1:].lower() in extensions:
        raise argparse.ArgumentTypeError(f"'{arg}' is not a file of type {' or '.join(extensions)}")
    return arg


def is_valid_dir(arg):
    if not os.path.isdir(arg):
        raise argparse.ArgumentTypeError(f"'{arg}' is not an existing directory, and I am afraid to create it")
    return arg


def check_species(arg):
    if not arg in speciesopts:
        raise argparse.ArgumentTypeError(f"'{arg}' is not a recognised species" + "\nUse gplas with the --speciesopts flag for a list of all supported species")
    return arg


def check_output(path):
    if not os.path.exists(path):
        print('\n')
        print(f"Failed to create the following output: {path}")
        quit_tool(-1)


def delete_file(file_path):
    if os.path.exists(file_path):
        os.remove(file_path)


def delete_empty_dir(dir_path):
    if os.path.exists(dir_path):
        if not any(os.listdir(dir_path)):
            os.rmdir(dir_path)


def cleanup_centrifuge(sample):
    delete_file(f"plasmidCC/{sample}/{sample}_plasmids.fasta")
    delete_file(f"plasmidCC/{sample}/{sample}_centrifuge_classified.txt")


def cleanup_intermediary_files(sample):
    #Coverage files
    delete_file(f"coverage/{sample}_clean_links.tab")
    delete_file(f"coverage/{sample}_clean_prediction.tab")
    delete_file(f"coverage/{sample}_clean_repeats.tab")
    delete_file(f"coverage/{sample}_estimation.txt")
    delete_file(f"coverage/{sample}_graph_contigs.tab")
    delete_file(f"coverage/{sample}_initialize_nodes.tab")
    delete_file(f"coverage/{sample}_isolated_nodes.tab")
    delete_file(f"coverage/{sample}_repeat_nodes.tab")
    delete_file(f"coverage/{sample}_repeats_graph.tab")
    #Walks normal mode
    delete_file(f"walks/normal_mode/{sample}_solutions.tab")
    #Walks bold mode + unbinned solutions
    delete_file(f"walks/bold_mode/{sample}_solutions_bold.tab")
    delete_file(f"walks/unbinned_nodes/{sample}_solutions_unbinned.tab")
    delete_file(f"walks/{sample}_solutions.tab")
    #Walks repeats
    delete_file(f"walks/repeats/{sample}_solutions.tab")
    #Results no_repeats
    delete_file(f"results/{sample}_results_no_repeats.tab")
    delete_file(f"results/{sample}_bins_no_repeats.tab")
    #Centrifuge classification
    delete_file(f"plasmidCC/{sample}/{sample}_gplas.tab")
    #Delete directories if they exist and are empty
    delete_empty_dir("coverage/")
    delete_empty_dir("walks/normal_mode/")
    delete_empty_dir("walks/bold_mode/")
    delete_empty_dir("walks/unbinned_nodes/")
    delete_empty_dir("walks/repeats/")
    delete_empty_dir("walks/")
    delete_empty_dir(f"plasmidCC/{sample}/")
    delete_empty_dir("plasmidCC/")



import pandas as pd
import statistics
from Bio.SeqIO.FastaIO import SimpleFastaParser

def coverage(sample, path_prediction, pred_threshold):
    #Inputs    
    path_nodes = f"gplas_input/{sample}_raw_nodes.fasta"
    path_links = f"gplas_input/{sample}_raw_links.txt"
    
    #Outputs
    output_graph_contigs = f"coverage/{sample}_graph_contigs.tab"
    output_clean_links = f"coverage/{sample}_clean_links.tab"
    output_graph_repeats = f"coverage/{sample}_repeats_graph.tab"
    output_clean_prediction = f"coverage/{sample}_clean_prediction.tab"
    output_isolated_nodes = f"coverage/{sample}_isolated_nodes.tab"
    output_clean_repeats = f"coverage/{sample}_clean_repeats.tab"
    output_initialize_nodes = f"coverage/{sample}_initialize_nodes.tab"
    output_repeat_nodes = f"coverage/{sample}_repeat_nodes.tab"
    output_cov_estimate = f"coverage/{sample}_estimation.txt"
        
    with open(path_nodes) as file:
        raw_nodes = [[str(values[0]), str(values[1])] for values in SimpleFastaParser(file)]
    
    raw_contig_names = [str(entry[0]) for entry in raw_nodes]
    
    kc_check = sum([name.count('KC') for name in raw_contig_names])
    
    if kc_check == len(raw_contig_names):
        lengths = [len(entry[1]) for entry in raw_nodes]
        kc_counts = [name.split(':', maxsplit=4)[2] for name in raw_contig_names]
        kc_counts = [int(name.replace('_','')) for name in kc_counts]
        kc_coverage = [kc/length for kc, length in zip(kc_counts, lengths)]
        coverage = [coverage/statistics.median(kc_coverage) for coverage in kc_coverage]
    else:
        raw_lengths = [name.split(':')[2] for name in raw_contig_names]
        lengths = [int(name.replace('_dp','')) for name in raw_lengths]
        coverage = [float(name.split(':')[4]) for name in raw_contig_names]

    raw_number = [name.split('_')[0] for name in raw_contig_names]
    number = [name.replace('S','') for name in raw_number]

    contig_info = pd.DataFrame(data={'number':number,
                                     'length':lengths,
                                     'coverage':coverage,
                                     'Contig_name':raw_contig_names})

    graph_pos_contigs = pd.DataFrame(data={'number':[digit+'+' for digit in number],
                                           'length':lengths,
                                           'coverage':coverage,
                                           'Contig_name':raw_contig_names})

    graph_neg_contigs = pd.DataFrame(data={'number':[digit+'-' for digit in number],
                                           'length':lengths,
                                           'coverage':coverage,
                                           'Contig_name':raw_contig_names})

    graph_contigs = pd.concat([graph_pos_contigs, graph_neg_contigs], ignore_index=True)
    graph_contigs.to_csv(output_graph_contigs, sep='\t', index=False, mode='w')

    raw_links = pd.read_csv(path_links, sep='\t', header=None)
    raw_links.rename({0:'L',1:'first_node',2:'first_sign',3:'second_node',4:'second_sign',5:'OM'}, axis='columns', inplace=True)

    links = []
    for index, row in raw_links.iterrows():
        reverse_first_sign = '+' if row['first_sign'] == '-' else '-'
        reverse_second_sign = '+' if row['second_sign'] == '-' else '-'

        clean_first = str(row['first_node'])+str(row['first_sign'])
        clean_second = str(row['second_node'])+str(row['second_sign'])
        info_forward = [clean_first, 'to', clean_second]
        links.append(info_forward)

        clean_rev_first = str(row['second_node'])+str(reverse_second_sign)
        clean_rev_second = str(row['first_node'])+str(reverse_first_sign)
        info_reverse = [clean_rev_first, 'to', clean_rev_second]
        links.append(info_reverse)

    links = pd.DataFrame(links)
    links.to_csv(output_clean_links, sep='\t', index=False, header=False, mode='w')

    unique_nodes = sorted(list(set(links[0])))

    outdegree_info = []
    for node in unique_nodes:
        repeat_links = links[links[0] == node]
        unique_links = repeat_links.drop_duplicates()
        outdegree_info.append([node, len(unique_links[2])])

    outdegree_info = pd.DataFrame(outdegree_info, columns=['number', 'outdegree'])

    indegree_info = []
    for node in unique_nodes:
        repeat_links = links[links[2] == node]
        unique_links = repeat_links.drop_duplicates()
        indegree_info.append([node, len(unique_links[0])])

    indegree_info = pd.DataFrame(indegree_info, columns=['number', 'indegree'])

    repeat_info = pd.merge(outdegree_info, indegree_info, on='number')
    repeats = repeat_info[(repeat_info['indegree'] > 1) | (repeat_info['outdegree'] > 1)]

    repeats.to_csv(output_graph_repeats, sep='\t', index=False, mode='w')

    repeats.loc[:,'number'] = [name.replace('+','') for name in repeats['number']]
    repeats.loc[:,'number'] = [name.replace('-','') for name in repeats['number']]

    pred = pd.read_table(path_prediction, sep='\t', header=0)

    raw_number = [name.split('_', maxsplit=1)[0] for name in pred['Contig_name']]
    pred.loc[:,'number'] = [number.replace('S','') for number in raw_number]
    pred = pd.merge(pred, contig_info, on=['Contig_name','number'])

    final_prediction = pred[[number not in list(repeats['number']) for number in pred['number']]]

    final_prediction.to_csv(output_clean_prediction, sep='\t', index=False, mode='w')

    unique_nodes_signless = links[0]
    unique_nodes_signless = [node.replace('+','') for node in unique_nodes_signless]
    unique_nodes_signless = [node.replace('-','') for node in unique_nodes_signless]

    isolated_nodes = contig_info[[number not in unique_nodes_signless for number in contig_info['number']]]
    isolated_nodes = pd.merge(pred, isolated_nodes['Contig_name'], on='Contig_name')
    isolated_nodes = isolated_nodes[isolated_nodes['Prob_Plasmid'] >= pred_threshold]

    isolated_nodes.to_csv(output_isolated_nodes, sep='\t', index=False, mode='w')

    repeats_final = pred[[number in list(repeats['number']) for number in pred['number']]]

    repeats_final.to_csv(output_clean_repeats, sep='\t', index=False, mode='w')

    pl_nodes = final_prediction[final_prediction['Prob_Plasmid'] >= pred_threshold]
    pl_nodes = pl_nodes[pl_nodes['Contig_length'] > 500]  # TODO ASK Julian: is this threshold needed?
    index = [number not in list(repeats['number']) for number in pl_nodes['number']]
    pl_nodes = pl_nodes.loc[index,:]

    initialize_nodes = sorted(list(set(pl_nodes['number'])))
    with open(output_initialize_nodes, 'w') as file:
        for node in initialize_nodes:
            file.write(node + '\n')

    repeats_nodes = sorted(list(set(repeats_final['number'])))
    with open(output_repeat_nodes, 'w') as file:
        for node in repeats_nodes:
            file.write(node + '\n')

    chr_contigs = pred[(pred['Prob_Chromosome'] > 0.7) & (pred['Contig_length'] > 1000)]  # TODO this is hardcoded, both values could/should become parameters?
    cov_estimation = chr_contigs[[number not in list(repeats['number']) for number in chr_contigs['number']]]
    sd_estimation = statistics.stdev(cov_estimation['coverage'])

    with open(output_cov_estimate, 'w') as file:
        file.write(str(sd_estimation))

    return



import pandas as pd
import numpy as np
import scipy.stats


def generate_paths(sample, number_iterations, filtering_threshold, sd_coverage=1, mode='normal'):
    #Inputs
    path_links = f"coverage/{sample}_clean_links.tab"
    path_prediction = f"coverage/{sample}_clean_prediction.tab"
    path_graph_contigs = f"coverage/{sample}_graph_contigs.tab"
    path_graph_repeats = f"coverage/{sample}_repeats_graph.tab"
    path_init_nodes = f"coverage/{sample}_initialize_nodes.tab"
    path_cov_variation = f"coverage/{sample}_estimation.txt"
    #Params
    # TODO make these into user-tunable parameters?
    number_nodes = 100
    prob_small_repeats = 0.5
    #Outputs Normal Mode
    if mode == 'normal':
        output_path = f"walks/normal_mode/{sample}_solutions.tab"
    #Outputs Bold Mode
    elif mode == 'bold':
        output_path = f"walks/bold_mode/{sample}_solutions_bold.tab"

    links = pd.read_csv(path_links, sep='\t', header=None)
    clean_pred = pd.read_csv(path_prediction, sep='\t', header=0)
    clean_pred = clean_pred.astype({'Prob_Chromosome':float,
                                    'Prob_Plasmid':float,
                                    'Prediction':str,
                                    'Contig_name':str,
                                    'Contig_length':int,
                                    'number':str,
                                    'length':int,
                                    'coverage':float})

    graph_contigs = pd.read_csv(path_graph_contigs, sep='\t', header=0)

    small_contigs = graph_contigs[graph_contigs['length'] < 500].copy()
    small_contigs_signed_nodes = small_contigs.copy()
    small_contigs_signed_nodes = [node for node in small_contigs_signed_nodes['number']]
    small_contigs.loc[:,'number'] = [name.replace('+','') for name in small_contigs['number']]
    small_contigs.loc[:,'number'] = [name.replace('-','') for name in small_contigs['number']]

    repeats = pd.read_csv(path_graph_repeats, sep='\t', header=0)
    repeats_signed_nodes = repeats.copy()
    repeats_signed_nodes = [node for node in repeats_signed_nodes['number']]
    repeats.loc[:,'number'] = [name.replace('+','') for name in repeats['number']]
    repeats.loc[:,'number'] = [name.replace('-','') for name in repeats['number']]

    initialize_nodes = pd.read_csv(path_init_nodes, sep='\t', header=None)
    initialize_nodes = [str(node) for node in initialize_nodes.iloc[:,0]]
    
    with open(path_cov_variation, mode='r') as file:
        max_variation = float(file.readline()) * sd_coverage


    def plasmid_graph(initial_seed, output_path, links, number_iterations, number_nodes, max_variation, filtering_threshold, prob_small_repeats, direction):
        paths_list = []
        for iteration in range(number_iterations): # Number of times we repeat this process
            path = [initial_seed] # We add the initial seed to the path, first element in the list
            seed = initial_seed

            #################################### Coverage of the current path #################################

            # Extracting the info from our current path                     
            info_path = graph_contigs[[number in path for number in graph_contigs['number']]].copy()
            info_path = info_path[[number not in repeats_signed_nodes for number in info_path['number']]] # Removing the contigs corresponding to transposases
            length_path = sum(info_path['length']) # Length of the path
            info_path['contribution'] = info_path['length']/length_path # Shorter contigs should have a lower contribution to the coverage. k-mer coverage on these contigs fluctuates drastically
            path_mean = np.average(info_path['coverage'], weights=info_path['contribution']) # Coverage of the current path

            ##################### Elongating the path ###########################################

            for elongation in range(number_nodes):
                index = links.loc[:,0] == seed
                current_links = links.loc[index,:] # Consider the last element present in our path and observe all the possible links

                if(current_links.shape[0] == 0):
                    output = '\t'.join(path)
                    paths_list.append(output)
                    path = [initial_seed] # There are no connections possible from this contig
                    break # Exiting elongation loop

                list_connections = sorted(list(set(current_links[2]))) # All the possible unique connections 

                # We do not allow that a node which is not a repeat appears more than 1 time in any solution but we exclude the initial seed from this consideration 
                if(len(path) > 1): # If the path has more than one element        
                    remove_nodes = path[1:]

                    remove_nodes = [node.replace('+','') for node in remove_nodes]
                    remove_nodes = [node.replace('-','') for node in remove_nodes]

                    positive_remove_nodes = [node + '+' for node in remove_nodes]
                    negative_remove_nodes = [node + '-' for node in remove_nodes]

                    ommit_nodes = positive_remove_nodes + negative_remove_nodes

                    first_node = path[0]

                    if(direction == 'forward'):
                        first_node_to_exclude = first_node.replace('+','-')

                    elif(direction == 'reverse'):
                        first_node_to_exclude = first_node.replace('-','+')

                    ommit_nodes.append(first_node_to_exclude)

                    # We need to remove directionality from the path to avoid paths e.g. 54-,161-,54+
                    ommit_nodes = [node for node in ommit_nodes if node not in repeats_signed_nodes]

                    list_connections = [str(connection) for connection in list_connections if connection not in ommit_nodes] # Avoiding loops within the solution

                total_connections = len(list_connections) # Number of connections
                base_probabilities = [float(0)]*total_connections # Generating a list with the connection probabilities

                if(total_connections < 1):
                    output = '\t'.join(path)
                    paths_list.append(output)
                    path = [initial_seed] # There are no connections possible from this contig
                    break # Exiting elongation loop

                ################################# Probabilities of the connections #####################################################

                # We generate a dataframe with the connection probabilities
                prob_df = pd.DataFrame(data={'number':list_connections,
                                             'Prob_Plasmid':base_probabilities})

                # Replacing the base probabilities with the probabilities generated by mlplasmids 
                prob_df.loc[:,'number'] = [number.replace('+','') for number in prob_df['number']]
                prob_df.loc[:,'number'] = [number.replace('-','') for number in prob_df['number']]

                for number in prob_df.loc[:,'number']:
                    matchID = clean_pred.index[clean_pred['number'] == number]
                    if any(matchID):
                        index = prob_df.loc[:,'number'] == number
                        prob_df.loc[index, 'Prob_Plasmid'] = float(clean_pred.loc[matchID, 'Prob_Plasmid'].values[0])

                # Short contigs do not have a reliable probability, we assign them a predefined probability (passed via the argument 'prob_small_repeats')
                index = prob_df.loc[:,'number'].isin(small_contigs.loc[:,'number'])
                prob_df.loc[index, 'Prob_Plasmid'] = float(prob_small_repeats) # OVERLAP!  

                # Transposases are also corner-cases for the machine-learning algorithm, we follow the same principle as done with short contigs 
                index = prob_df.loc[:,'number'].isin(repeats.loc[:,'number'])
                prob_df.loc[index, 'Prob_Plasmid'] = float(prob_small_repeats)

                final_probs = list(prob_df.loc[:,'Prob_Plasmid'])

                record_connections = []
                for i in range(len(final_probs)):
                    record_connections.append([1.0, number_iterations, iteration, elongation, initial_seed, seed, list_connections[i], final_probs[i]])

                record_connections = pd.DataFrame(data=record_connections, columns=['factor','number_iterations','iteration','elongation','first_node','ingoing_node','outgoing_node','Probability_pl_chr'])

                index = graph_contigs.loc[:,'number'].isin(record_connections.loc[:,'outgoing_node'])
                cov_connections_info = graph_contigs.loc[index,:].copy()

                up_cutoff = cov_connections_info.loc[:,'coverage'].copy() + max_variation            
                down_cutoff = cov_connections_info.loc[:,'coverage'].copy() - max_variation

                up_threshold = scipy.stats.norm.cdf(up_cutoff, loc=path_mean, scale=max_variation)
                down_threshold = scipy.stats.norm.cdf(down_cutoff, loc=path_mean, scale=max_variation)

                # Simple test
                window = up_threshold - down_threshold
                cov_connections_info.loc[:,'Probability_cov'] = abs(window)
                #TODO find a better way to do this merge
                record_connections = record_connections.merge(cov_connections_info[['number','Probability_cov']], left_on='outgoing_node', right_on='number')
                record_connections = record_connections.drop(columns='number')

                index = record_connections.loc[:,'outgoing_node'].isin(repeats_signed_nodes)
                record_connections.loc[index,'Probability_cov'] = float(prob_small_repeats)

                index = record_connections.loc[:,'outgoing_node'].isin(small_contigs_signed_nodes)
                record_connections.loc[index,'Probability_cov'] = float(prob_small_repeats)                        

                record_connections.loc[:,'Probability'] = record_connections.loc[:,'Probability_pl_chr'] * record_connections.loc[:,'Probability_cov']
                record_connections.loc[:,'Probability_freq'] = record_connections.loc[:,'Probability']/sum(record_connections.loc[:,'Probability'])
                record_connections.loc[:,'Verdict'] = 'non-selected'

                if(sum(record_connections.loc[:,'Probability'] >= filtering_threshold) == 0):
                    output = '\t'.join(path)
                    paths_list.append(output)
                    path = [initial_seed] # There are no connections possible from this contig
                    break # Exiting elongation loop

                # Filter step to avoid going into really bad connections
                index = list(record_connections.loc[:,'Probability'] >= filtering_threshold)
                filter_connections = record_connections.loc[index,:].copy()
                if(sum(filter_connections['Probability_freq']) != 1): # recalculate probability frequencies after filter
                    filter_connections['Probability_freq'] = filter_connections.loc[:,'Probability']/sum(filter_connections.loc[:,'Probability'])

                random_connection = np.random.choice(filter_connections.loc[:,'outgoing_node'], size=1, p=filter_connections.loc[:,'Probability_freq'])[0] # Choose one connection 
                path.append(str(random_connection))

                if(random_connection == path[0]):
                    output = '\t'.join(path)
                    paths_list.append(output)
                    path = [initial_seed] 
                    break # Exiting elongation loop

                seed = str(random_connection)

                if(len(path) >= number_nodes): # We only exit the function if we have reached the maximum number of nodes allowed per path
                    output = '\t'.join(path)
                    paths_list.append(output)
                    path = [initial_seed] 
                    break # Exiting elongation loop

                index = graph_contigs.loc[:,'number'].isin(path)
                info_path = graph_contigs.loc[index,:].copy()

                index = ~info_path.loc[:,'number'].isin(repeats_signed_nodes)
                info_path = info_path.loc[index,:]

                length_path = sum(info_path.loc[:,'length'])
                info_path.loc[:,'contribution'] = info_path.loc[:,'length'] / length_path
                path_mean = np.average(info_path.loc[:,'coverage'], weights=info_path.loc[:,'contribution']) # Coverage of the current path

        return paths_list


    final_paths = []
    for seed in initialize_nodes:
        np.random.seed(123)
        positive_seed = seed + '+'
        negative_seed = seed + '-'
        final_paths.extend(plasmid_graph(positive_seed, output_path, links, number_iterations, number_nodes, max_variation, filtering_threshold, prob_small_repeats, direction='forward'))
        final_paths.extend(plasmid_graph(negative_seed, output_path, links, number_iterations, number_nodes, max_variation, filtering_threshold, prob_small_repeats, direction='reverse'))

    with open(output_path, mode='w') as outfile:
        for path in final_paths:
            outfile.write(path + '\n')

    return



from setuptools import setup, find_packages

with open("requirements.txt") as file_open:
     requirements = file_open.read().splitlines()

with open("README.md") as file_open:
     README = file_open.read()

setup(
    name="gplas",
    setup_requires=[
        "setuptools>=38.6.0",
        "setuptools_scm",
        "setuptools_scm_git_archive",
    ],
    use_scm_version={"version_file":"gplas/version.py"},
    #version="1.0.0",
    description="Binning plasmid-predicted contigs using short-read graphs",
    long_description=README,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    entry_points={
        'console_scripts': ["gplas = gplas.gplas:main"],
    }

)
